package com.ruoyi.project.business.service.flow.impl;

import com.alibaba.fastjson2.JSON;
import com.alibaba.fastjson2.JSONObject;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.ruoyi.common.exception.ServiceErrorException;
import com.ruoyi.common.utils.SecurityUtils;
import com.ruoyi.framework.security.LoginUser;
import com.ruoyi.project.business.domain.flow.*;
import com.ruoyi.project.business.domain.system.SysAttachment;
import com.ruoyi.project.business.dto.TaskGenerateParams;
import com.ruoyi.project.business.enums.*;
import com.ruoyi.project.business.mapper.flow.FlowConfigMapper;
import com.ruoyi.project.business.mapper.flow.FlowMapper;
import com.ruoyi.project.business.mapper.flow.FlowTaskMapper;
import com.ruoyi.project.business.mapper.flow.FlowTaskUserMapper;
import com.ruoyi.project.business.service.flow.IFlowBusinessService;
import com.ruoyi.project.business.service.flow.IFlowService;
import com.ruoyi.project.business.service.system.ISysAttachmentService;
import com.ruoyi.project.system.domain.SysRole;
import com.ruoyi.project.system.domain.SysUser;
import com.ruoyi.project.system.service.ISysRoleService;
import com.ruoyi.project.system.service.ISysUserService;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.util.*;
import java.util.function.Function;

import static com.ruoyi.project.business.enums.AuditStatusEnum.*;
import static com.ruoyi.project.business.enums.SignTypeEnum.COUNTER_SIGN;
import static java.util.stream.Collectors.*;
import static org.apache.commons.lang3.StringUtils.isBlank;
import static org.apache.commons.lang3.StringUtils.isNotBlank;

@Service
public class FlowServiceImpl implements IFlowService {

    private static final Logger log = LoggerFactory.getLogger(FlowServiceImpl.class);

    @Resource
    private FlowConfigMapper flowConfigMapper;
    @Resource
    private FlowMapper flowMapper;
    @Resource
    private FlowTaskMapper flowTaskMapper;
    @Resource
    private FlowTaskUserMapper flowTaskUserMapper;

    @Resource
    private ISysUserService sysUserService;

    @Resource
    private ISysRoleService sysRoleService;

    @Resource
    private IFlowBusinessService flowBusinessService;

    @Resource
    private ISysAttachmentService sysAttachmentService;


    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean rollback(FlowAuditParam auditParam) {
        Flow flow = flowMapper.getBusinessFlow(auditParam.getAuditFlow(), auditParam.getBusinessId());
        if (Objects.isNull(flow) || !AUDITOR_ONE.getValue().equals(flow.getStatus())) {
            throw new ServiceErrorException("业务数据状态已变更，请刷新后重试");
        }
        flow.setStatus(ROLLBACK.getValue());
        flow.setUpdateTime(new Date());
        flow.setUpdateBy(SecurityUtils.getUserId().toString());
        flowMapper.updateStatus(flow);
        auditParam.setStatus(ROLLBACK.getValue());
        auditParam.setFinish(true);
        auditParam.setPromoter(flow.getPromoter());
        auditParam.setBusinessData(flow.getBusinessData());
        flowBusinessService.changeBusinessStatus(auditParam);
        return true;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean auditBatch(FlowAuditParam auditParam) {
        List<Long> businessIds = auditParam.getBusinessIds();
        if (CollectionUtils.isEmpty(businessIds)) {
            throw new ServiceErrorException("请选择审批数据");
        }
        businessIds.forEach(businessId -> {
            FlowAuditParam param = new FlowAuditParam();
            BeanUtils.copyProperties(auditParam, param);
            param.setBusinessId(businessId);
            this.audit(param);
        });
        return true;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean audit(FlowAuditParam auditParam) {
        AuditStatusEnum auditStatus = AuditStatusEnum.from(auditParam.getStatus());
        Flow flow = flowMapper.getBusinessFlow(auditParam.getAuditFlow(), auditParam.getBusinessId());
        if (Objects.isNull(flow)) {
            throw new ServiceErrorException("数据异常，审批流程数据不存在");
        }
        auditParam.setBusinessData(flow.getBusinessData());
        if (Objects.isNull(flow) || !AuditStatusEnum.isAuditing(flow.getStatus())) {
            throw new ServiceErrorException("业务数据状态已变更，请刷新后重试");
        }
        auditParam.setPromoter(flow.getPromoter());
        LoginUser currentUser = SecurityUtils.getLoginUser();
        FlowTask currentTask = flowTaskMapper.getFlowTasks(flow.getId()).stream()
                .filter(flowTask -> flowTask.getStatus().equals(AUDITING.getValue())).findFirst().get();
        List<FlowTaskUser> currentTaskUsers = flowTaskUserMapper.listTaskUsers(Lists.newArrayList(currentTask.getId()))
                .stream().filter(user -> user.getStatus().equals(AUDITING.getValue())).collect(toList());

        // 1. 记录用户审批结果
        FlowTaskUser param = new FlowTaskUser();
        param.setStatus(auditParam.getStatus());
        param.setRemark(auditParam.getRemark());
        param.setAuditor(SecurityUtils.getUserId().toString());
        param.setAuditTime(new Date());
        param.setUpdateTime(new Date());
        param.setUpdateBy(SecurityUtils.getUserId().toString());

        param.setIds(currentTaskUsers.stream().filter(flowTaskUser ->
                        flowTaskUser.getAuditUser().equals(currentUser.getUserId().toString()) && isAuditing(flowTaskUser.getStatus()))
                .map(FlowTaskUser::getId).collect(toList()));

        if (CollectionUtils.isEmpty(param.getIds())) {
            throw new ServiceErrorException("当前用户不具备审批权限");
        }

        flowTaskUserMapper.updateStatusBatch(param);
        sysAttachmentService.updateAttachments(param.getIds().get(0).intValue(),
                ModuleEnum.from(auditParam.getModule()),
                FunctionEnum.from(auditParam.getFunc()),
                CategoryEnum.from(auditParam.getCategory()),
                auditParam.getAttachmentList());

        // 如果是驳回直接中断审批
        if (REJECT.equals(auditStatus)) {
            currentTask.setStatus(REJECT.getValue());
            currentTask.setUpdateTime(new Date());
            currentTask.setUpdateBy(currentUser.getUsername());
            flowTaskMapper.updateStatus(currentTask);
            flow.setStatus(REJECT.getValue());
            flow.setUpdateTime(new Date());
            flow.setUpdateBy(currentUser.getUsername());
            flowMapper.updateStatus(flow);
            auditParam.setFinish(true);
            flowBusinessService.changeBusinessStatus(auditParam);
            return true;
        }

        // 会签存在人未审批则本次处理结束
        SignTypeEnum signType = SignTypeEnum.from(currentTask.getSignType());
        if (COUNTER_SIGN.equals(signType) &&
                flowTaskUserMapper.listTaskUsers(Lists.newArrayList(currentTask.getId()))
                        .stream().filter(user -> user.getStatus().equals(AUDITING.getValue())).count() > 0) {
            return true;
        }

        // 当前节点审批通过
        currentTask.setStatus(AuditStatusEnum.PASS.getValue());
        currentTask.setUpdateBy(SecurityUtils.getUserId().toString());
        currentTask.setUpdateTime(new Date());
        flowTaskMapper.updateStatus(currentTask);

        // 无后置节点则流程结束
        String nodeId = currentTask.getNextTasks();
        if (isBlank(nodeId)) {
            flow.setStatus(AuditStatusEnum.PASS.getValue());
            flow.setUpdateTime(new Date());
            flow.setUpdateBy(currentUser.getUsername());
            flowMapper.updateStatus(flow);
            auditParam.setFinish(true);
            auditParam.setBusinessData(flow.getBusinessData());
            flowBusinessService.changeBusinessStatus(auditParam);
            return true;
        }

        // 激活后置节点
        FlowTask nextTask = flowTaskMapper.getNext(flow.getId(), nodeId);
        nextTask.setStatus(AUDITING.getValue());
        nextTask.setUpdateTime(new Date());
        nextTask.setUpdateBy(currentUser.getUsername());
        flowTaskMapper.updateStatus(nextTask);

        flow.setStatus(nextTask.getStatusCode());
        flow.setUpdateTime(new Date());
        flow.setUpdateBy(currentUser.getUsername());
        flowMapper.updateStatus(flow);

        List<FlowTaskUser> nextTaskUsers = flowTaskUserMapper.listTaskUsers(Lists.newArrayList(nextTask.getId()));
        FlowTaskUser nextParam = new FlowTaskUser();
        nextParam.setIds(nextTaskUsers.stream().map(FlowTaskUser::getId).collect(toList()));
        nextParam.setStatus(AUDITING.getValue());
        nextParam.setUpdateTime(new Date());
        nextParam.setUpdateBy(currentUser.getUsername());
        flowTaskUserMapper.updateStatusBatch(nextParam);

        auditParam.setStatus(nextTask.getStatusCode());
        auditParam.setNextAuditUsers(nextTaskUsers.stream().map(FlowTaskUser::getAuditUser).map(Long::valueOf).collect(toList()));
        flowBusinessService.changeBusinessStatus(auditParam);

        return true;
    }


    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean generateTasks(TaskGenerateParams taskGenerateParams) {
        AuditFlowEnum auditFlow = taskGenerateParams.getAuditFlow();
        JSONObject jsonObject = JSON.parseObject(JSON.toJSONString(taskGenerateParams.getBusinessData()));
        if (Objects.isNull(jsonObject.get("id")) || Objects.isNull(taskGenerateParams.getUserId())) {
            throw new RuntimeException("业务数据异常，缺少业务数据id或申请人id");
        }

        Long businessId = Long.parseLong(jsonObject.get("id").toString());
        Long userId = taskGenerateParams.getUserId();

        Flow exist = getFlowMap(Lists.newArrayList(businessId.intValue()), auditFlow).get(businessId.intValue());
        if (Objects.nonNull(exist) && AuditStatusEnum.isAuditing(exist.getStatus())) {
            throw new ServiceErrorException("已存在进行中的审批，请勿重复提交");
        }
        // 删除当前业务下的所有任务
        remove(auditFlow, Lists.newArrayList(businessId));

        SysUser promoter = sysUserService.selectUserById(userId);
        Flow flow = new Flow();
        flow.setBusinessId(businessId);
        flow.setBusinessDesc(auditFlow.getName() + "-" + promoter.getUserName());
        flow.setAuditFlow(auditFlow.getValue());
        flow.setPromoter(userId);
        flow.setBusinessData(jsonObject.toString());
        flow.setCreateTime(new Date());
        flow.setCreateBy(SecurityUtils.getUserId().toString());

        // 1. 根据审批数据直接所属公司及流程类型获取审批流程图
        FlowConfig param = new FlowConfig();
        param.setCompanyId(1L);
        param.setAuditFlow(auditFlow.getValue());
        FlowConfig flowConfig = getConfig(param);
        NodeConfig nodeConfig = null;
        List<NodeConfig> auditors = null;
        if (Objects.nonNull(flowConfig)) {
            nodeConfig = JSON.parseObject(flowConfig.getFlowData(), NodeConfig.class);
            auditors = extractAuditors(flow, nodeConfig);
        }
        if (Objects.isNull(nodeConfig) || CollectionUtils.isEmpty(auditors)) {
            flow.setFlowConfigId(-1L);
            flow.setStatus(AuditStatusEnum.PASS.getValue());
            flow.setAuditFlow(auditFlow.getValue());
            flowMapper.save(flow);
            FlowAuditParam auditParam = new FlowAuditParam();
            auditParam.setStatus(AuditStatusEnum.PASS.getValue());
            auditParam.setBusinessId(flow.getBusinessId());
            auditParam.setBusinessData(flow.getBusinessData());
            auditParam.setAuditFlow((long) flow.getAuditFlow());
            auditParam.setFinish(true);
            flowBusinessService.changeBusinessStatus(auditParam);
            return false;
        } else {
            //审核中
            flow.setStatus(AUDITING.getValue());
        }
        // 2. 生成并保存流程及其任务节点信息
        flow.setFlowConfigId(flowConfig.getId());
        flow.setAuditFlow(flowConfig.getAuditFlow());
        flowMapper.save(flow);
        // 3. 根据条件生成审批流程(路径中没有审批人则直接通过)
        Long flowId = flow.getId();
        if (CollectionUtils.isEmpty(auditors)) {
            Flow auditParam = new Flow();
            auditParam.setId(flowId);
            auditParam.setStatus(AuditStatusEnum.PASS.getValue());
            flowMapper.updateStatus(auditParam);
            FlowAuditParam params = new FlowAuditParam();
            params.setStatus(AuditStatusEnum.PASS.getValue());
            params.setBusinessId(flow.getBusinessId());
            params.setBusinessData(flow.getBusinessData());
            params.setAuditFlow((long) flow.getAuditFlow());
            params.setFinish(true);
            flowBusinessService.changeBusinessStatus(params);
            return false;
        }
        // 3.1 设置前后置关系及状态
        NodeConfig prevNode = null;
        for (NodeConfig node : auditors) {
            if (Objects.nonNull(prevNode)) {
                node.setPrevTasks(prevNode.getNodeId());
                prevNode.setNextTasks(node.getNodeId());
                node.setStatus(DRAFT.getValue());
            } else {
                node.setStatus(AUDITING.getValue());
            }
            prevNode = node;
        }
        // 3.2 以审批人为单位生成任务
        List<FlowTask> flowTaskList = Lists.newArrayList();
        List<FlowTaskUser> flowTaskUserList = Lists.newArrayList();
        List<Long> roleIds = auditors.stream()
                .map(NodeConfig::getRoles)
                .flatMap(Collection::stream)
                .map(SysRole::getRoleId).distinct().collect(toList());
        Map<Long, List<Long>> roleUserMap = sysRoleService.getRoleUserMap(roleIds);

        for (NodeConfig node : auditors) {
            FlowTask flowTask = new FlowTask();
            BeanUtils.copyProperties(node, flowTask);
            flowTask.setFlowId(flowId);
            flowTask.setNodeName(node.getNodeType().getLabel());
            flowTask.setStatusCode(Integer.valueOf(node.getNodeType().getValue()));
            flowTask.setSignType(Integer.valueOf(node.getSignType().getValue()));
            flowTask.setAuditorType(Integer.valueOf(node.getAuditorType().getValue()));
            flowTaskList.add(flowTask);
        }
        flowTaskMapper.saveBatch(flowTaskList);

        Map<String, Long> idMap = flowTaskList.stream().collect(toMap(FlowTask::getNodeId, FlowTask::getId));
        for (NodeConfig node : auditors) {
            AuditorTypeEnum auditorType = AuditorTypeEnum.from(Integer.valueOf(node.getAuditorType().getValue()));
            List<Long> userIds = Lists.newArrayList();
            switch (auditorType) {
                case USER:
                    userIds = node.getUsers().stream()
                            .map(SysUser::getUserId).collect(toList());
                    break;
                case ROLE:
                    userIds = node.getRoles().stream()
                            .map(sysRole -> roleUserMap.get(sysRole.getRoleId()))
                            .flatMap(Collection::stream).collect(toList());
                    break;
            }
            // 将申请人从审批人中移除
            userIds.remove(SecurityUtils.getUserId());
            flowTaskUserList.addAll(userIds.stream().distinct().map(id -> {
                FlowTaskUser flowTaskUser = new FlowTaskUser();
                flowTaskUser.setTaskId(idMap.get(node.getNodeId()));
                flowTaskUser.setStatus(node.getStatus());
                flowTaskUser.setAuditUser(id.toString());
                flowTaskUser.setCreateTime(new Date());
                return flowTaskUser;
            }).collect(toList()));
            flowTaskUserList.stream().forEach(flowTaskUser -> {
                flowTaskUser.setCreateTime(new Date());
            });
        }
        flowTaskUserMapper.saveBatch(flowTaskUserList);
        List<Long> auditorUserIds = flowTaskUserList.stream()
                .filter(flowtaskUser -> AUDITING.getValue().equals(flowtaskUser.getStatus()))
                .map(FlowTaskUser::getAuditUser).map(Long::valueOf).collect(toList());

        FlowAuditParam params = new FlowAuditParam();
        params.setStatus(flowTaskList.get(0).getStatusCode());
        params.setBusinessId(flow.getBusinessId());
        params.setBusinessData(flow.getBusinessData());
        params.setAuditFlow((long) flow.getAuditFlow());
        params.setNextAuditUsers(auditorUserIds);
        flowBusinessService.changeBusinessStatus(params);
        flow.setStatus(params.getStatus());
        flowMapper.updateStatus(flow);
        return true;
    }

    private List<NodeConfig> extractAuditors(Flow flow, NodeConfig nodeConfig) {
        List<NodeConfig> auditNodes = Lists.newArrayList();

        NodeConfig nextNode = nodeConfig.getChildNode();
        while (Objects.nonNull(nextNode)) {
            NodeTypeEnum nodeType = NodeTypeEnum.from(nextNode.getType());
            switch (nodeType) {
                case AUDITOR:
                    auditNodes.add(nextNode);
                    NodeConfig child = nextNode.getChildNode();
                    NodeConfig parent = nextNode.getParent();

                    if (Objects.nonNull(child)) {
                        child.setParent(parent);
                        nextNode = child;
                    } else if (Objects.nonNull(parent)) {
                        // 如果是子流程结束需要回到父流程继续
                        nextNode = parent.getChildNode();
                    } else {
                        nextNode = null;
                    }
                    break;
            }
        }

        return auditNodes;
    }

    @Override
    public int remove(AuditFlowEnum auditFlow, List<Long> businessIds) {
        return flowMapper.remove(auditFlow, businessIds);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public int saveConfig(FlowConfig flowConfig) {
        FlowConfig exist = getConfig(flowConfig);
        if (Objects.nonNull(exist)) {
            flowConfigMapper.remove(flowConfig);
        }
        return flowConfigMapper.save(flowConfig);
    }

    @Override
    public FlowConfig getConfig(FlowConfig flowConfig) {
        return flowConfigMapper.get(flowConfig.getCompanyId(), flowConfig.getAuditFlow());
    }

    @Override
    public int removeConfig(FlowConfig flowConfig) {
        return flowConfigMapper.remove(flowConfig);
    }

    @Override
    public Map<Integer, Flow> getFlowMap(List<Integer> ids, AuditFlowEnum auditFlow) {
        return getFlowMap(ids, Lists.newArrayList(auditFlow));
    }

    @Override
    public Map<Integer, Flow> getFlowMap(List<Integer> ids, List<AuditFlowEnum> auditFlows) {
        if (CollectionUtils.isEmpty(ids)) {
            return Maps.newHashMap();
        }
        List<Flow> flowList = flowMapper.getFlowList(ids, auditFlows, AUDITING.getValue());
        if (CollectionUtils.isEmpty(flowList)) {
            return Maps.newHashMap();
        }
        List<Long> taskIds = flowList.stream().map(Flow::getCurrentTaskId).collect(toList());
        Map<Integer, Flow> result = Maps.newHashMap();
        Map<Long, List<FlowTaskUser>> flowUserMap = flowTaskUserMapper.listTaskUsers(taskIds).stream().collect(groupingBy(FlowTaskUser::getTaskId));

        flowList.forEach(flow -> {
            flow.setCurrentAuditors(flowUserMap.get(flow.getCurrentTaskId()));
            Integer key = flow.getBusinessId().intValue();
            Flow exist = result.get(key);
            if (Objects.isNull(exist) || exist.getCreateTime().before(flow.getCreateTime())) {
                result.put(key, flow);
            }
        });

        return result;
    }

    @Override
    public List<FlowTaskUser> getAuditHistory(List<AuditFlowEnum> auditFlows, Integer businessId, ModuleEnum module, FunctionEnum func, CategoryEnum category) {
        if (CollectionUtils.isEmpty(auditFlows)) {
            throw new ServiceErrorException("流程类型不能为空");
        }
        if (Objects.isNull(businessId)) {
            throw new ServiceErrorException("业务数据id不能为空");
        }
        List<Flow> flowList = flowMapper.getAuditHistoryFlowList(auditFlows, (long) businessId);
        if (CollectionUtils.isEmpty(flowList)) {
            return Lists.newArrayList();
        }
        List<FlowTask> flowTaskList = flowTaskMapper.getHistoryFlowTasks(flowList.stream().map(Flow::getId).collect(toList()));
        if (CollectionUtils.isEmpty(flowTaskList)) {
            return Lists.newArrayList();
        }
        List<FlowTaskUser> flowTaskUserList = flowTaskUserMapper.listTaskUsers(flowTaskList.stream().map(FlowTask::getId).collect(toList()));
        if (CollectionUtils.isEmpty(flowTaskUserList)) {
            return Lists.newArrayList();
        }

        List<FlowTaskUser> historyList = flowTaskUserList.stream()
                .filter(flowTaskUser -> Objects.nonNull(flowTaskUser.getAuditTime()))
                .sorted(Comparator.comparing(FlowTaskUser::getAuditTime)).collect(toList());
        Map<Integer, List<SysAttachment>> attachmentMap = sysAttachmentService.getAttachmentsMap(
                historyList.stream().map(flowTaskUser -> flowTaskUser.getId().intValue()).collect(toList()), module, func, category);

        Map<Long, FlowTask> taskMap = flowTaskList.stream().collect(toMap(FlowTask::getId, Function.identity()));
        historyList.forEach(history -> {
            FlowTask task = taskMap.get(history.getTaskId());
            history.setFlowId(task.getFlowId());
            history.setAttachmentList(attachmentMap.get(history.getId().intValue()));
            String auditor = history.getAuditorName();
            if (!history.getAuditUser().equals(history.getAuditor()) && isNotBlank(auditor)) {
                history.setNickName(history.getNickName() + "(" + auditor + "代签)");
            }
            history.setDesc(task.getNodeName() + (PASS.getValue().equals(history.getStatus()) ? "通过" : "驳回"));
        });
        Map<Long, List<FlowTaskUser>> histroyMap = historyList.stream().collect(groupingBy(FlowTaskUser::getFlowId));
        List<FlowTaskUser> result = Lists.newArrayList();
        flowList.forEach(flow -> {
            // 1. 添加提交节点
            FlowTaskUser submit = new FlowTaskUser();
            submit.setDesc(flow.getPromoterName() + "提交[" + AuditFlowEnum.from(flow.getAuditFlow()).getName() + "]");
            submit.setAuditTime(flow.getCreateTime());
            result.add(submit);

            // 2. 添加撤回节点
            if (ROLLBACK.getValue().equals(flow.getStatus())) {
                FlowTaskUser rollback = new FlowTaskUser();
                rollback.setDesc(flow.getPromoterName() + "撤回");
                rollback.setAuditTime(flow.getUpdateTime());
                result.add(rollback);
            }

            // 2. 添加审批节点
            List<FlowTaskUser> auditList = histroyMap.get(flow.getId());
            if (CollectionUtils.isNotEmpty(auditList)) {
                result.addAll(auditList);
            }
        });
        return result;
    }
}
