package com.ruoyi.project.business.service.quotation.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.date.DateUtil;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ruoyi.common.enums.SerialNumberPrefix;
import com.ruoyi.common.utils.SecurityUtils;
import com.ruoyi.common.utils.bean.BeanUtils;
import com.ruoyi.framework.redis.RedisCache;
import com.ruoyi.project.business.domain.TcProject;
import com.ruoyi.project.business.domain.TqSpecification;
import com.ruoyi.project.business.domain.TqSpecificationDetail;
import com.ruoyi.project.business.domain.TqSpecificationParabarDetail;
import com.ruoyi.project.business.domain.bo.quotation.ElevatorAddBO;
import com.ruoyi.project.business.domain.bo.quotation.ElevatorParamAddBO;
import com.ruoyi.project.business.domain.bo.quotation.QuotationAddBO;
import com.ruoyi.project.business.domain.bo.quotation.QuotationCopyBO;
import com.ruoyi.project.business.domain.bo.quotation.QuotationQueryBO;
import com.ruoyi.project.business.domain.vo.quotation.QuotationDetailVO;
import com.ruoyi.project.business.domain.vo.quotation.QuotationVO;
import com.ruoyi.project.business.domain.vo.quotation.QuotationVersionVO;
import com.ruoyi.project.business.domain.vo.selectParam.ProjectParam;
import com.ruoyi.project.business.enums.DelFlagEnum;
import com.ruoyi.project.business.enums.ParamTypeEnum;
import com.ruoyi.project.business.mapper.TcProjectMapper;
import com.ruoyi.project.business.mapper.quotation.TqSpecificationMapper;
import com.ruoyi.project.business.mapper.quotation.TqSpecificationParabarDetailMapper;
import com.ruoyi.project.business.service.quotation.ITqSpecificationDetailService;
import com.ruoyi.project.business.service.quotation.ITqSpecificationParabarDetailService;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import com.ruoyi.common.utils.DateUtils;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import com.ruoyi.project.business.mapper.quotation.TqQuotationMapper;
import com.ruoyi.project.business.domain.TqQuotation;
import com.ruoyi.project.business.service.quotation.ITqQuotationService;
import org.springframework.transaction.annotation.Transactional;

/**
 * 【请填写功能名称】Service业务层处理
 *
 * @author ruoyi
 * @date 2024-05-10
 */
@Service
@Slf4j
public class TqQuotationServiceImpl extends ServiceImpl<TqQuotationMapper, TqQuotation> implements ITqQuotationService {

    @Resource
    private TqSpecificationMapper tqSpecificationMapper;

    @Resource
    private TqSpecificationParabarDetailMapper tqSpecificationParabarDetailMapper;

    @Resource
    private TcProjectMapper tcProjectMapper;

    @Resource
    private ITqSpecificationParabarDetailService iTqSpecificationParabarDetailService;

    @Resource
    private RedisCache redisCache;

    @Resource
    private ITqSpecificationDetailService iTqSpecificationDetailService;

    /**
     * 查询【请填写功能名称】
     *
     * @param id 【请填写功能名称】主键
     * @return 【请填写功能名称】
     */
    @Override
    public QuotationDetailVO selectTqQuotationById(Long id) {
        QuotationDetailVO quotationDetailVO = baseMapper.selectDetailById(id);
        ProjectParam tcProject = new ProjectParam();
        tcProject.setProjectId(quotationDetailVO.getProjectId());
        tcProject.setPermissionChar(SecurityUtils.getPermissionCharSea());
        List<TcProject> tcProjects = tcProjectMapper.selectTcProjectList(tcProject);
        if (CollUtil.isNotEmpty(tcProjects)) {
            BeanUtils.copyProperties(tcProjects.get(0), quotationDetailVO, "currencyId");
        }
        return quotationDetailVO;
    }

    /**
     * 查询【请填写功能名称】列表
     *
     * @param queryBO 【请填写功能名称】
     * @return 【请填写功能名称】
     */
    @Override
    public List<QuotationVO> selectTqQuotationList(QuotationQueryBO queryBO) {
        Optional.ofNullable(queryBO.getCreateTime()).ifPresent(item -> {
            if (item.size() > 1) {
                item.set(1, DateUtil.endOfDay(item.get(1)));
            }
        });
        List<QuotationVO> quotationVOS = baseMapper.selectTqQuotationList(queryBO);

        // 查询子项
        List<String> quotationNoList = quotationVOS.stream().map(QuotationVO::getQuotationNo)
            .collect(Collectors.toList());
        List<Long> quotatiionIdList = quotationVOS.stream().map(QuotationVO::getId)
            .collect(Collectors.toList());
        if (CollUtil.isNotEmpty(quotationNoList)) {
            QuotationQueryBO quotationQueryBO = new QuotationQueryBO();
            quotationQueryBO.setQuotationNoList(quotationNoList);
            quotationQueryBO.setExcludeQuotationIdList(quotatiionIdList);
            quotationQueryBO.setIsGroupBy(false);
            List<QuotationVO> childrenList = baseMapper.selectTqQuotationList(quotationQueryBO);
            if (CollUtil.isNotEmpty(childrenList)) {
                Map<String, List<QuotationVO>> childrenMap = childrenList.stream()
                    .collect(Collectors.groupingBy(QuotationVO::getQuotationNo));
                for (QuotationVO quotationVO : quotationVOS) {
                    quotationVO.setChildren(childrenMap.getOrDefault(quotationVO.getQuotationNo(),
                        Collections.emptyList()));
                }
            }
        }
        return quotationVOS;
    }

    /**
     * 新增【请填写功能名称】
     *
     * @param quotationAddBO 【请填写功能名称】
     * @return 结果
     */
    @Override
    @Transactional
    public void insertTqQuotation(QuotationAddBO quotationAddBO) {
        // 报价
        TqQuotation tqQuotation = new TqQuotation();
        tqQuotation.setQuotationNo(redisCache.generateSerialNumber(SerialNumberPrefix.BJ.getPrefix()));
        tqQuotation.setVersion(1L);
        tqQuotation.setAgentId(quotationAddBO.getAgentId());
        tqQuotation.setProjectId(quotationAddBO.getProjectId());
        tqQuotation.setCurrencyId(quotationAddBO.getCurrencyId());
        tqQuotation.setSubmitTime(new Date());
        // 数据权限字符
        tqQuotation.setPermissionChar("default");
        tqQuotation.setQuotationStatus(quotationAddBO.getQuotationStatus());
        tqQuotation.setDelFlag(DelFlagEnum.NORMAL.getValue());
        tqQuotation.setCreateBy(SecurityUtils.getStrUserId());
        tqQuotation.setCreateTime(DateUtils.getNowDate());
        tqQuotation.setUpdateBy(SecurityUtils.getStrUserId());
        tqQuotation.setUpdateTime(DateUtils.getNowDate());
        baseMapper.insertTqQuotation(tqQuotation);
        Long quotationId = tqQuotation.getId();
        log.info("报价id [{}]", quotationId);

        insertSpecificationParam(quotationAddBO, quotationId);
    }

    // 新增电梯规则 和 参数
    private void insertSpecificationParam(QuotationAddBO quotationAddBO, Long quotationId) {
        for (ElevatorAddBO elevator : quotationAddBO.getElevatorList()) {
            TqSpecification tqSpecification = new TqSpecification();
            tqSpecification.setQuotationId(quotationId);
            tqSpecification.setElevatorModelId(elevator.getElevatorModelId());
            tqSpecification.setElevatorNumber(elevator.getElevatorNumber());
            tqSpecification.setSpecification(elevator.getSpecification());
            tqSpecification.setQuotationStatus(quotationAddBO.getQuotationStatus());
            tqSpecification.setDelFlag(DelFlagEnum.NORMAL.getValue());
            tqSpecification.setCreateBy(SecurityUtils.getStrUserId());
            if (Objects.nonNull(elevator.getCreateTime())) {
                tqSpecification.setCreateTime(elevator.getCreateTime());
            } else {
                tqSpecification.setCreateTime(DateUtils.getNowDate());
            }
            tqSpecification.setUpdateBy(SecurityUtils.getStrUserId());
            tqSpecification.setUpdateTime(DateUtils.getNowDate());
            tqSpecificationMapper.insertTqSpecification(tqSpecification);
            Long specificationId = tqSpecification.getId();
            log.info("规格id [{}]", quotationId);
            // 过滤出来普通参数
            List<ElevatorParamAddBO> normalParamList = elevator.getParamList().stream()
                .filter(item -> ParamTypeEnum.NORMAL.getCode().equals(item.getParamType())).collect(
                    Collectors.toList());
            for (ElevatorParamAddBO paramAddBO : normalParamList) {
                TqSpecificationDetail tqSpecificationDetail = new TqSpecificationDetail();
                tqSpecificationDetail.setSpecificationId(specificationId);
                tqSpecificationDetail.setParaCode(paramAddBO.getParaCode());
                tqSpecificationDetail.setParaValue(paramAddBO.getParaValue());
                tqSpecificationDetail.setParaCategory(paramAddBO.getParaCategory());
                tqSpecificationDetail.setDelFlag(DelFlagEnum.NORMAL.getValue());
                tqSpecificationDetail.setCreateBy(SecurityUtils.getStrUserId());
                tqSpecificationDetail.setCreateTime(DateUtils.getNowDate());
                tqSpecificationDetail.setUpdateBy(SecurityUtils.getStrUserId());
                tqSpecificationDetail.setUpdateTime(DateUtils.getNowDate());
                iTqSpecificationDetailService.insertTqSpecificationDetail(tqSpecificationDetail);
            }
            // 过滤参数包参数
            List<ElevatorParamAddBO> parabarParamList = elevator.getParamList().stream()
                .filter(item -> ParamTypeEnum.BAR_PARAM.getCode().equals(item.getParamType())).collect(
                    Collectors.toList());
            for (ElevatorParamAddBO paramAddBO : parabarParamList) {
                TqSpecificationDetail tqSpecificationDetail = new TqSpecificationDetail();
                tqSpecificationDetail.setSpecificationId(specificationId);
                tqSpecificationDetail.setParaCode(paramAddBO.getParaCode());
                tqSpecificationDetail.setParaValue(paramAddBO.getParaValue());
                tqSpecificationDetail.setParabarId(paramAddBO.getParabarId());
                tqSpecificationDetail.setParabarDetailId(paramAddBO.getParabarDetailId());
                tqSpecificationDetail.setParaCategory(paramAddBO.getParaCategory());
                tqSpecificationDetail.setDelFlag(DelFlagEnum.NORMAL.getValue());
                tqSpecificationDetail.setCreateBy(SecurityUtils.getStrUserId());
                tqSpecificationDetail.setCreateTime(DateUtils.getNowDate());
                tqSpecificationDetail.setUpdateBy(SecurityUtils.getStrUserId());
                tqSpecificationDetail.setUpdateTime(DateUtils.getNowDate());
                iTqSpecificationDetailService.insertTqSpecificationDetail(tqSpecificationDetail);
            }
        }
    }

    /**
     * 修改【请填写功能名称】
     *
     * @param quotationAddBO 【请填写功能名称】
     * @return 结果
     */
    @Override
    @Transactional
    public void updateTqQuotation(QuotationAddBO quotationAddBO) {
        tqSpecificationMapper.deleteByQuotationId(quotationAddBO.getQuotationId());
        tqSpecificationParabarDetailMapper.deleteByQuotationId(quotationAddBO.getQuotationId());
        insertSpecificationParam(quotationAddBO, quotationAddBO.getQuotationId());
    }

    /**
     * 批量删除【请填写功能名称】
     *
     * @param ids 需要删除的【请填写功能名称】主键
     * @return 结果
     */
    @Override
    public int deleteTqQuotationByIds(Long[] ids) {
        return baseMapper.deleteTqQuotationByIds(ids);
    }

    /**
     * 删除【请填写功能名称】信息
     *
     * @param id 【请填写功能名称】主键
     * @return 结果
     */
    @Override
    public int deleteTqQuotationById(Long id) {
        return baseMapper.deleteTqQuotationById(id);
    }

    @Override
    @Transactional
    public void copyVersions(QuotationCopyBO copyBO) {
        List<TqQuotation> tqQuotations = baseMapper.selectBatchIds(copyBO.getIds());
        // 获取报价编号的最大版本号
        List<QuotationVersionVO> maxVersionById = baseMapper.getMaxVersionById(copyBO.getIds());
        Map<String, Long> quotationVersionMap = maxVersionById.stream()
            .collect(Collectors.toMap(QuotationVersionVO::getQuotationNo, QuotationVersionVO::getVersion, (a, b) -> a));
        // 报价复制版本
        for (TqQuotation tqQuotation : tqQuotations) {
            Long oldQuotationId = tqQuotation.getId();
            tqQuotation.setVersion(quotationVersionMap.getOrDefault(tqQuotation.getQuotationNo(), 1L) + 1L);
            baseMapper.insertTqQuotation(tqQuotation);
            Long newQuotationId = tqQuotation.getId();
            List<TqSpecification> tqSpecifications = tqSpecificationMapper.selectList(
                Wrappers.<TqSpecification>lambdaQuery().eq(TqSpecification::getQuotationId, oldQuotationId)
            );
            tqSpecifications.forEach(item -> item.setQuotationId(newQuotationId));
            // 电梯规格复制版本
            for (TqSpecification tqSpecification : tqSpecifications) {
                Long oldSpecificationId = tqSpecification.getId();
                tqSpecificationMapper.insertTqSpecification(tqSpecification);
                Long newSpecificationId = tqSpecification.getId();
                List<TqSpecificationParabarDetail> paramDetail = tqSpecificationParabarDetailMapper.selectList(
                    Wrappers.<TqSpecificationParabarDetail>lambdaQuery()
                        .eq(TqSpecificationParabarDetail::getSpecificationDetailId, oldSpecificationId)
                );
                // 电梯规格参数复制版本
                paramDetail.forEach(item -> {
                    item.setSpecificationDetailId(newSpecificationId);
                    item.setId(null);
                });
                iTqSpecificationParabarDetailService.saveBatch(paramDetail);
            }
        }
    }
}
